#include <stdio.h>
#include <string.h>
#include <mpi.h>

#define M 4
#define N 4

void PrintMatrix(double *matrix){
  for(int i=0;i<M;i++){
    for(int j=0;j<N;j++){
      printf("%02.0lf ",matrix[i*N+j]);
    }
    printf("\n");
  }
}

void PrintVector(double *vector){
  for(int j=0;j<N;j++){
    printf("%02.0lf ",vector[j]);
  }
  printf("\n");
}


void MatrixVectorMultiply(
                          double *local_matrix, // The block distributed version of the matrix
                          double *local_vect,   // The block distributed version of the vector
                          double *result,       // Will contains one block of the result for the current process
                          double local_m,       // Number of row per process
                          double local_n        // Number of vector component per process
                          ){
  // Will contains the vector used for the multiplication
  double vect[N];

  // First gather the vector from other process (from the previous multiplication)
  MPI_Allgather(local_vect,local_n,MPI_DOUBLE, vect, local_n,MPI_DOUBLE,MPI_COMM_WORLD);

  // Perform the local dot product
  for(int i=0;i<local_m;i++){
    result[i]=0.0;
    for(int j=0;j<N;j++){
      result[i]+=local_matrix[i*N+j]*vect[j];
    }
  }
}


int main(){
    int nproc; // Number of process running the distributed program (-n argument of mpiexec)
    int myrank; // Rank of the current process

    /* Define a matrix
      0  1  2  3
      4  5  6  7
      8  9  10 11
      12 13 14 15
     */
    double matrix[M*N]={0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15};
    double vect[N]={1,2,3,4};
    
    MPI_Init(NULL,NULL);
    MPI_Comm_size(MPI_COMM_WORLD,&nproc);
    MPI_Comm_rank(MPI_COMM_WORLD,&myrank);

    // Print infos
    if(myrank==0){
      printf("----- Matrix -----\n");
      PrintMatrix(matrix);
      printf("----- Vector -----\n");
      PrintVector(vect);
    }
    
    // Init local variables
    int local_m=M/nproc; // Number of row per process
    int local_n=N/nproc; // Number of column == Number of components in vector vect == number of components per process
    double local_vect[N]; // Contains the local part of vect
    double local_matrix[M*N]; // Contains the local matrix

    // Distribute the matrix rows in blocks
    MPI_Scatter(matrix, local_m*N, MPI_DOUBLE, local_matrix, local_m*M, MPI_DOUBLE, 0, MPI_COMM_WORLD);

    // Distribute the vector components in blocks
    // Indeed, since MatrixVectorMultiply fetch the vector from a gather we should distribute it
    MPI_Scatter(vect, local_n, MPI_DOUBLE, local_vect, local_n, MPI_DOUBLE, 0, MPI_COMM_WORLD);

    // Perform multiplication
    double result[N]; // To store the result
    MatrixVectorMultiply(local_matrix, local_vect, result, local_m, local_n);
    
    // Fetch (in vect) and print the result
    MPI_Gather(result, local_n, MPI_DOUBLE, vect,local_n,MPI_DOUBLE,0,MPI_COMM_WORLD);
    if(myrank==0){
      printf("----- Result -----\n");
      PrintVector(vect);
    }
    
    // Now if you want to do another multiplication with the the same matrix and the resulting vector
    /* MatrixVectorMultiply(local_matrix, result, result, local_m, local_n); */
    /* MPI_Gather(result, local_n, MPI_DOUBLE, vect,local_n,MPI_DOUBLE,0,MPI_COMM_WORLD); */
    /* if(myrank==0) */
    /*   PrintVector(vect); */

    
    MPI_Finalize();

    return(0);
}
